#!/usr/bin/env python3
# coding: utf-8
"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""
from sense.client.requestwrapper import RequestWrapper
from sense.common import classwrapper

@classwrapper
class LoggingApi():
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """
    def __init__(self, req_wrapper=None):
        if req_wrapper is None:
            self.client = RequestWrapper()
        else:
            self.client = req_wrapper
        if 'SI_UUID' in self.client.config:
            self.si_uuid = self.client.config['SI_UUID']
        else:
            self.si_uuid = None

    def logging_set_archive_days(self, days, **kwargs):  # noqa: E501
        """Archive logs  # noqa: E501

        Archives all logs older than specified number of days.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_set_archive_days(days, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param float days: Number of days. (required)
        :return: str
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.logging_archive_days_put_with_http_info(
                days, **kwargs)  # noqa: E501
        else:
            (data) = self.logging_archive_days_put_with_http_info(
                days, **kwargs)  # noqa: E501
            return data

    def logging_archive_days_put_with_http_info(self, days,
                                                **kwargs):  # noqa: E501
        """Archive logs  # noqa: E501

        Archives all logs older than specified number of days.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_archive_days_put_with_http_info(days, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param float days: Number of days. (required)
        :return: str
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['days']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method logging_archive_days_put" % key)
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'days' is set
        if ('days' not in params or params['days'] is None):
            raise ValueError(
                "Missing the required parameter `days` when calling `logging_archive_days_put`"
            )  # noqa: E501

        path_params = {}
        if 'days' in params:
            path_params['days'] = params['days']  # noqa: E501

        return self.client.request('PUT', f'/logging/archive/{days}')

    def logging_get_config(self, **kwargs):  # noqa: E501
        """Get logging configuration  # noqa: E501

        Retrieves the logging levels for each layer as well as database usage metadata.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_get_config(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :return: LoggingConfiguration
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.logging_config_get_with_http_info(**
                                                          kwargs)  # noqa: E501
        else:
            (data
             ) = self.logging_config_get_with_http_info(**kwargs)  # noqa: E501
            return data

    def logging_config_get_with_http_info(self, **kwargs):  # noqa: E501
        """Get logging configuration  # noqa: E501

        Retrieves the logging levels for each layer as well as database usage metadata.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_config_get_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :return: LoggingConfiguration
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = []  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method logging_config_get" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('GET', f'/logging/config')

    def logging_set_logger_level(self, logger, level, **kwargs):  # noqa: E501
        """Set logger level  # noqa: E501

        Updates the logging level of the specified layer.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_set_logger_level(logger, level, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str logger: Logging appender layer. (required)
        :param str level: Logging appender level. (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.logging_config_logger_level_put_with_http_info(
                logger, level, **kwargs)  # noqa: E501
        else:
            (data) = self.logging_config_logger_level_put_with_http_info(
                logger, level, **kwargs)  # noqa: E501
            return data

    def logging_config_logger_level_put_with_http_info(self, logger, level,
                                                       **kwargs):  # noqa: E501
        """Set logger level  # noqa: E501

        Updates the logging level of the specified layer.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_config_logger_level_put_with_http_info(logger, level, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str logger: Logging appender layer. (required)
        :param str level: Logging appender level. (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['logger', 'level']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method logging_config_logger_level_put" %
                                key)
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'logger' is set
        if ('logger' not in params or params['logger'] is None):
            raise ValueError(
                "Missing the required parameter `logger` when calling `logging_config_logger_level_put`"
            )  # noqa: E501
        # verify the required parameter 'level' is set
        if ('level' not in params or params['level'] is None):
            raise ValueError(
                "Missing the required parameter `level` when calling `logging_config_logger_level_put`"
            )  # noqa: E501

        path_params = {}
        if 'logger' in params:
            path_params['logger'] = params['logger']  # noqa: E501
        if 'level' in params:
            path_params['level'] = params['level']  # noqa: E501

        return self.client.request('PUT', f'/logging/config/{logger}/{level}')

    def logging_set_filter(self, **kwargs):  # noqa: E501
        """Set logger filter  # noqa: E501

        Updates the logging filter with a new configuration string.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_set_filter(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str body: Configuration string. "Use raw" will filter over the raw logging message before processing, and "pattern" is the regex itself.
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.logging_filter_put_with_http_info(**
                                                          kwargs)  # noqa: E501
        else:
            (data
             ) = self.logging_filter_put_with_http_info(**kwargs)  # noqa: E501
            return data

    def logging_filter_put_with_http_info(self, **kwargs):  # noqa: E501
        """Set logger filter  # noqa: E501

        Updates the logging filter with a new configuration string.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_filter_put_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str body: Configuration string. "Use raw" will filter over the raw logging message before processing, and "pattern" is the regex itself.
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method logging_filter_put" % key)
            params[key] = val
        del params['kwargs']

        body_params = None
        if 'body' in params:
            body_params = params['body']

        return self.client.request('PUT',
                                   f'/logging/filter',
                                   body_params=body_params)

    def instance_get_logging(self, **kwargs):  # noqa: E501
        """Retrieve intents by service instance  # noqa: E501

        Queries all service intents belonging to given instance UUID.  # noqa: E501
        This method makes a synchronous HTTP request by default.
        :param async_req bool
        :param str si_uuid: Intent UUID. (required)
        :return: list[IntentExpanded]
                 If the method is called asynchronously,
                 returns the request thread.
        """
        if self.si_uuid:
            kwargs['si_uuid'] = self.si_uuid
        if not kwargs['si_uuid'] :
            raise ValueError("Missing the required parameter `si_uuid`")

        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.logging_logs_si_uuid_get_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.logging_logs_si_uuid_get_with_http_info(**kwargs)  # noqa: E501
            return data

    def logging_logs_si_uuid_get_with_http_info(self,
                                                **kwargs):  # noqa: E501
        """Retrieve intents by service instance  # noqa: E501

        Queries all service intents belonging to given instance UUID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.logging_logs_si_uuid_get_with_http_info(si_uuid, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str si_uuid: Intent UUID. (required)
        :return: list[IntentExpanded]
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['si_uuid']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method logging_logs_si_uuid_get_with_http_info" % key)
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'si_uuid' is set
        if ('si_uuid' not in params or params['si_uuid'] is None):
            raise ValueError(
                "Missing the required parameter `si_uuid` when calling `logging_logs_si_uuid_get_with_http_info`"
            )  # noqa: E501

        return self.client.request('GET', '/logging/logs/' + kwargs['si_uuid'])



import sys

from sense.workflow.base import utils
from sense.workflow.base import state_utils as sutil
from sense.workflow.base.config import WorkflowConfig
from sense.workflow.base.exceptions import ControllerException
from sense.workflow.controller import Controller


def delete_session_if_empty(*, session):
    provider_states = sutil.load_states(session)

    if not provider_states:
        sutil.destroy_session(session)


def manage_workflow(args):
    logger = utils.init_logger()
    config_dir = utils.absolute_path(args.config_dir)
    sessions = sutil.load_sessions()
    config_dir_from_meta = sutil.load_meta_data(args.session, 'config_dir') if args.session in sessions else None

    if config_dir_from_meta and config_dir_from_meta != config_dir:
        logger.error(f"attempt to use session {args.session} from the wrong config dir {config_dir} ...")
        logger.warning(f"ATTN: The CORRECT config dir for session {args.session} is {config_dir_from_meta}!!!!!!!")
        sys.exit(1)

    var_dict = utils.load_vars(args.var_file) if args.var_file else {}

    if args.validate:
        try:
            WorkflowConfig.parse(dir_path=config_dir, var_dict=var_dict)
            logger.info("config looks ok")
        except Exception as e:
            logger.error(f"Validation failed .... {type(e)} {e}")
            logger.error(e, exc_info=True)
            sys.exit(1)

    if args.plan:
        config = WorkflowConfig.parse(dir_path=config_dir, var_dict=var_dict)
        controller = Controller(config=config)
        states = sutil.load_states(args.session)
        controller.init(session=args.session, provider_states=states)
        controller.plan(provider_states=states)
        cr, dl = sutil.dump_plan(resources=controller.resources, to_json=args.json, summary=args.summary)

        logger.warning(f"Applying this plan would create {cr} resource(s) and destroy {dl} resource(s)")
        delete_session_if_empty(session=args.session)
        return

    if args.apply:
        sutil.save_meta_data(dict(config_dir=config_dir), args.session)
        config = WorkflowConfig.parse(dir_path=config_dir, var_dict=var_dict)

        try:
            controller = Controller(config=config)
        except Exception as e:
            logger.error(f"Exceptions while initializing controller .... {e}", exc_info=True)
            sys.exit(1)

        states = sutil.load_states(args.session)

        try:
            controller.init(session=args.session, provider_states=states)
        except Exception as e:
            logger.error(f"Exceptions while initializing providers  .... {e}", exc_info=True)
            sys.exit(1)

        try:
            controller.plan(provider_states=states)
        except Exception as e:
            logger.error(f"Exception while planning ... {e}")
            sys.exit(1)
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while planning ... {kie}")
            sys.exit(1)

        try:
            controller.add(provider_states=states)
        except Exception as e:
            logger.error(f"Exception while adding ... {e}")
            sys.exit(1)
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while adding  resources ... {kie}")
            sys.exit(1)

        workflow_failed = False

        try:
            controller.apply(provider_states=states)
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while creating resources ... {kie}")
            workflow_failed = True
        except ControllerException as ce:
            logger.error(f"Exceptions while creating resources ... {ce}")
            workflow_failed = True
        except Exception as e:
            logger.error(f"Unknown error while creating resources ... {e}")
            workflow_failed = True

        states = controller.get_states()
        services, pending, failed = utils.get_counters(states=states)
        workflow_failed = workflow_failed or pending or failed

        if workflow_failed:
            states = sutil.reconcile_states(states, args.session)

        sutil.save_states(states, args.session)
        logger.info(f"services={services}, pending={pending}, failed={failed}")
        sys.exit(1 if workflow_failed else 0)

    if args.show:
        states = sutil.load_states(args.session) if args.session in sessions else []
        sutil.dump_states(states, args.json, args.summary)
        return

    if args.destroy:
        if args.session not in sessions:
            return

        states = sutil.load_states(args.session)

        if not states:
            sutil.destroy_session(args.session)
            return

        config = WorkflowConfig.parse(dir_path=config_dir, var_dict=var_dict)

        try:
            controller = Controller(config=config)
            controller.init(session=args.session, provider_states=states)
        except Exception as e:
            logger.error(f"Exceptions while initializing controller .... {e}")
            sys.exit(1)

        destroy_failed = False

        try:
            controller.destroy(provider_states=states)
        except ControllerException as e:
            logger.error(f"Exceptions while deleting resources ...{e}")
            destroy_failed = True
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while deleting resources ... {kie}")
            sys.exit(1)

        if not states:
            logger.info(f"Destroying session {args.session} ...")
            sutil.destroy_session(args.session)
        else:
            sutil.save_states(states, args.session)

        sys.exit(1 if destroy_failed else 0)


def manage_sessions(args):
    if args.show:
        utils.dump_sessions(args.json)
        return


def main(argv=None):
    argv = argv or sys.argv[1:]
    parser = utils.build_parser(manage_workflow=manage_workflow,
                                manage_sessions=manage_sessions)
    args = parser.parse_args(argv)

    if len(args.__dict__) == 0:
        parser.print_usage()
        sys.exit(1)

    args.dispatch_func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



from collections import namedtuple
from typing import Dict, Set, Any

from .exceptions import ParseConfigException
from .constants import Constants


class BaseConfig:
    def __init__(self, type: str, name: str, attrs: Dict):
        self.type = type.lower()
        self._var_name = name.lower()
        self.attributes = attrs

    @property
    def label(self) -> str:
        return self._var_name + '@' + self.type

    @property
    def var_name(self) -> str:
        return self._var_name

    @property
    def name(self) -> str:
        return self.attributes['name'] if 'name' in self.attributes else self._var_name

    def attribute(self, key: str):
        return self.attributes.get(key)

    def __str__(self) -> str:
        return self.var_name + '@' + self.type

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return (self.var_name, self.type) == (other.var_name, other.type)

    def __hash__(self):
        return hash(str(self))


class Config(BaseConfig):
    def __init__(self, type: str, name: str, attrs: Dict):
        super().__init__(type, name, attrs)


class ProviderConfig(BaseConfig):
    def __init__(self, type: str, name: str, attrs: Dict):
        super().__init__(type, name, attrs)


Dependency = namedtuple("Dependency", "key resource attribute is_external")


class ResourceConfig(BaseConfig):
    def __init__(self, type: str, name: str, attrs: Dict, provider: ProviderConfig):
        super().__init__(type, name, attrs)
        assert provider, f"provider is required for {name}"
        assert isinstance(provider, ProviderConfig), f"expected ProviderConfig for {name}"
        self._provider = provider
        self._resource_dependencies = set()

    @property
    def provider(self) -> ProviderConfig:
        return self._provider

    @property
    def dependencies(self) -> Set[Dependency]:
        return self._resource_dependencies

    def add_dependency(self, dependency):
        self._resource_dependencies.add(dependency)

    def has_dependencies(self):
        return len(self._resource_dependencies) > 0

    @property
    def is_service(self):
        return self.type == Constants.RES_TYPE_SERVICE


def resource_from_basic_config(basic_config, providers) -> ResourceConfig:
    if isinstance(basic_config, ResourceConfig):
        return basic_config

    attrs = basic_config.attributes.copy()

    if 'provider' in attrs:
        provider = attrs.pop('provider')

        index = providers.index(ProviderConfig(provider.type, provider.var_name, {}))

        if index < 0:
            raise ParseConfigException(f"did not find provider {provider.var_name} for {basic_config}")
    else:
        provider = ProviderConfig("sense", basic_config.var_name + "_provider", dict())
        providers.append(provider)

    return ResourceConfig(basic_config.type, basic_config.var_name, attrs, provider)


DependencyInfo = namedtuple("DependencyInfo", "resource  attribute")


class Variable:
    def __init__(self, name: str, value: Any):
        self._name = name.lower()
        self._value = value

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> str:
        return self._value

    def __str__(self) -> str:
        return 'Variable[' + self.name + '=' + str(self.value) + ']'

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)



#!/usr/bin/env python3
# coding: utf-8
"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""
from sense.client.requestwrapper import RequestWrapper
from sense.common import classwrapper

@classwrapper
class IntentApi():
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """
    def __init__(self, req_wrapper=None):
        if req_wrapper is None:
            self.client = RequestWrapper()
        else:
            self.client = req_wrapper
        if 'SI_UUID' in self.client.config:
            self.si_uuid = self.client.config['SI_UUID']
        else:
            self.si_uuid = None

    def instance_get_intents(self, **kwargs):  # noqa: E501
        """Retrieve intents by service instance  # noqa: E501

        Queries all service intents belonging to given instance UUID.  # noqa: E501
        This method makes a synchronous HTTP request by default.
        :param async_req bool
        :param str si_uuid: Intent UUID. (required)
        :return: list[IntentExpanded]
                 If the method is called asynchronously,
                 returns the request thread.
        """
        if self.si_uuid:
            kwargs['si_uuid'] = self.si_uuid
        if not kwargs['si_uuid'] :
            raise ValueError("Missing the required parameter `si_uuid`")

        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.intent_instance_si_uuid_get_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.intent_instance_si_uuid_get_with_http_info(**kwargs)  # noqa: E501
            return data

    def intent_instance_si_uuid_get_with_http_info(self,
                                                   **kwargs):  # noqa: E501
        """Retrieve intents by service instance  # noqa: E501

        Queries all service intents belonging to given instance UUID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.intent_instance_si_uuid_get_with_http_info(si_uuid, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str si_uuid: Intent UUID. (required)
        :return: list[IntentExpanded]
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['si_uuid']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method intent_instance_si_uuid_get" % key)
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'si_uuid' is set
        if ('si_uuid' not in params or params['si_uuid'] is None):
            raise ValueError(
                "Missing the required parameter `si_uuid` when calling `intent_instance_si_uuid_get`"
            )  # noqa: E501

        return self.client.request('GET', '/intent/instance/' + kwargs['si_uuid'])

    def intent_describe(self, uuid, **kwargs):  # noqa: E501
        """Retrieve intent by UUID  # noqa: E501

        Queries service intent with given ID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.intent_describe(uuid, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str uuid: Intent UUID. (required)
        :return: Intent
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.intent_uuid_get_with_http_info(uuid,
                                                       **kwargs)  # noqa: E501
        else:
            (data) = self.intent_uuid_get_with_http_info(
                uuid, **kwargs)  # noqa: E501
            return data

    def intent_uuid_get_with_http_info(self, uuid, **kwargs):  # noqa: E501
        """Retrieve intent by UUID  # noqa: E501

        Queries service intent with given ID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.intent_uuid_get_with_http_info(uuid, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str uuid: Intent UUID. (required)
        :return: Intent
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['uuid']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method intent_uuid_get" % key)
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'uuid' is set
        if ('uuid' not in params or params['uuid'] is None):
            raise ValueError(
                "Missing the required parameter `uuid` when calling `intent_uuid_get`"
            )  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'uuid' in params:
            path_params['uuid'] = params['uuid']  # noqa: E501

        # Authentication setting
        auth_settings = ['oAuth2Keycloak']  # noqa: E501

        return self.client.request('GET', '/intent/' + uuid)



import logging
from abc import ABC, abstractmethod
from typing import List, Dict, Union

from sense.workflow.base.resource_models import Resource, Service
from sense.workflow.base.state_models import ProviderState, ServiceState
from sense.workflow.base.constants import Constants


class Provider(ABC):
    def __init__(self, *, type, label, name, logger: logging.Logger, config: dict):
        self.label = label
        self.type = type
        self.name = name
        self.logger = logger
        self.config = config
        self.resource_listener = None

        self._services = list()

        self._pending = []

        self._no_longer_pending = []
        self._failed = {}
        self.creation_details = {}
        self._added = []
        self.pending_internal = []

        self._saved_state: ProviderState = Union[ProviderState, None]
        self._existing_map: Dict[str, List[str]] = dict()
        self._added_map: Union[Dict[str, List[str]], None] = dict()

    @property
    def existing_map(self) -> Dict[str, List[str]]:
        return self._existing_map

    @property
    def added_map(self) -> Dict[str, List[str]]:
        return self._added_map

    @property
    def resources(self) -> List[Resource]:
        resources = []
        resources.extend([n for n in self._services])
        return resources

    @property
    def saved_state(self) -> Union[ProviderState, None]:
        return self._saved_state

    def set_saved_state(self, state: Union[ProviderState, None]):
        self._saved_state = state

    @property
    def services(self) -> List[Service]:
        return self._services

    @property
    def no_longer_pending(self) -> List:
        return self._no_longer_pending

    @property
    def pending(self) -> List:
        return self._pending

    @property
    def failed(self) -> Dict:
        return self._failed

    def set_resource_listener(self, resource_listener):
        self.resource_listener = resource_listener

    def on_added(self, *, source, provider, resource: object):
        assert self != source
        assert provider
        assert resource

    def on_deleted(self, *, source, provider, resource: object):
        assert self != source
        assert provider
        assert resource

    def get_dependency_resolver(self, *, external=True):
        from .dependency_reslover import DependencyResolver

        return DependencyResolver(label=self.label, external=external, logger=self.logger)

    def on_created(self, *, source, provider, resource: Resource):
        assert self != source
        assert provider

        if self == provider:
            self.creation_details[resource.label]["resources"].append(resource.name)
        else:
            for pending_resource in self.pending.copy():
                resolver = self.get_dependency_resolver()
                label = pending_resource[Constants.LABEL]
                resolver.resolve_dependency(resource=pending_resource, from_resource=resource)
                ok = resolver.check_if_dependencies_are_resolved(resource=pending_resource)

                if ok:
                    resolver.extract_values(resource=pending_resource)
                    self.pending.remove(pending_resource)
                    self.no_longer_pending.append(pending_resource)
                    self.logger.info(f"Removing {label} from pending using {self.label}")

    def init(self):
        credential_file = self.config.get(Constants.CREDENTIAL_FILE)
        from sense.common import getConfig

        if credential_file:
            self.config = getConfig(configFile=credential_file)
        else:
            self.config = getConfig()

        self.setup_environment()

    @abstractmethod
    def supports_modify(self):
        pass

    def resource_name(self, resource: dict, idx: int = 0):
        return f"{self.name}_{resource[Constants.RES_NAME_PREFIX]}_{idx}"

    def add_to_existing_map(self, resource: dict):
        label = resource.get(Constants.LABEL)
        self.existing_map[label] = []

        if self.saved_state and resource.get(Constants.LABEL) in self.saved_state.creation_details:
            provider_saved_creation_details = self.saved_state.creation_details[label]

            for n in range(0, provider_saved_creation_details['total_count']):
                resource_name = self.resource_name(resource, n)
                self.existing_map[label].append(resource_name)

                if label not in self.added_map:
                    self.added_map[label] = []

    def _compute_added_map(self):
        for service in self.services:
            if service.label not in self.added_map:
                self.added_map[service.label] = []

            self.added_map[service.label].append(service.name)

    @property
    def modified(self):
        if self.saved_state:
            self._compute_added_map()
            return self._existing_map and self.added_map != self._existing_map

        return False

    def validate_resource(self, *, resource: dict):
        label = resource.get(Constants.LABEL)
        self.creation_details[label] = dict()
        self.creation_details[label]['resources'] = list()
        self.creation_details[label]['config'] = resource[Constants.CONFIG]
        self.creation_details[label]['total_count'] = resource[Constants.RES_COUNT]
        self.creation_details[label]['failed_count'] = 0
        self.creation_details[label]['created_count'] = 0
        self.creation_details[label]['name_prefix'] = resource[Constants.RES_NAME_PREFIX]
        self.add_to_existing_map(resource)

        creation_details = resource[Constants.RES_CREATION_DETAILS]

        if not creation_details['in_config_file']:
            return

        try:
            self.do_validate_resource(resource=resource)
        except Exception as e:
            self.failed[label] = 'VALIDATE'
            raise e

    def add_resource(self, *, resource: dict):
        count = resource.get(Constants.RES_COUNT, 1)
        label = resource.get(Constants.LABEL)
        assert count > 0
        assert label not in self._added, f"{label} already in {self._added}"

        if len(resource[Constants.EXTERNAL_DEPENDENCIES]) > len(resource[Constants.RESOLVED_EXTERNAL_DEPENDENCIES]):
            self.logger.info(f"Adding {label} to pending using {self.label}")
            assert resource not in self.pending, f"Did not expect {label} to be in pending list using {self.label}"
            self.pending.append(resource)
            return
        elif len(resource[Constants.INTERNAL_DEPENDENCIES]) > len(resource[Constants.RESOLVED_INTERNAL_DEPENDENCIES]):
            self.logger.info(f"Handling internal dependencies {label} using provider {self.label}")
            resolver = self.get_dependency_resolver(external=False)

            for temp in self.resources:
                resolver.resolve_dependency(resource=resource, from_resource=temp)

            ok = resolver.check_if_dependencies_are_resolved(resource=resource)

            if ok:
                resolver.extract_values(resource=resource)
            else:
                self.logger.info(f"Adding to internal_dependencies {label}")

                assert resource not in self.pending_internal, f"internal pending resource {label} already added"
                self.pending_internal.append(resource)
                return

        try:
            creation_details = resource[Constants.RES_CREATION_DETAILS]

            if creation_details['in_config_file']:
                self.do_add_resource(resource=resource)

            self._added.append(label)
        except Exception as e:
            label = resource.get(Constants.LABEL)
            self.failed[label] = 'ADD'
            failed_count = resource[Constants.RES_COUNT] - len(self.creation_details[label]['resources'])
            self.creation_details[label]['failed_count'] = failed_count
            raise e

    def create_resource(self, *, resource: dict):
        label = resource.get(Constants.LABEL)

        if self.no_longer_pending:
            self.logger.info(f"Checking internal dependencies using {self.label}")
            temp_no_longer_pending = self._no_longer_pending
            self._no_longer_pending = []

            for no_longer_pending_resource in temp_no_longer_pending:
                external_dependency_label = no_longer_pending_resource[Constants.LABEL]

                try:
                    self.logger.info(f"Adding no longer pending external_dependency {external_dependency_label}")
                    self.add_resource(resource=no_longer_pending_resource)
                    added = True
                except Exception as e:
                    self.no_longer_pending.append(no_longer_pending_resource)
                    # Propagate the error only if the resourcce being created is not add successfully
                    if label == external_dependency_label:
                        raise e

                    self.logger.warning(f"Adding no longer pending externally {external_dependency_label} failed: {e}",
                                        exc_info=True)
                    added = False

                if added:
                    temp = self.pending_internal
                    self.pending_internal = []

                    for internal_dependency in temp:
                        internal_dependency_label = internal_dependency[Constants.LABEL]
                        self.logger.info(f"Adding internal_dependency {internal_dependency_label}")

                        try:
                            self.add_resource(resource=internal_dependency)
                        except Exception as e2:
                            self.logger.warning(
                                f"Adding no longer pending internally {internal_dependency_label} failed using {e2}",
                                exc_info=True)

        assert label in self._added
        self.logger.info(f"Create: {label} using {self.label}: {self._added}")

        try:
            self.do_create_resource(resource=resource)
        except (Exception, KeyboardInterrupt) as e:
            self.failed[label] = 'CREATE'
            failed_count = resource[Constants.RES_COUNT] - len(self.creation_details[label]['resources'])
            self.creation_details[label]['failed_count'] = failed_count
            raise e
        finally:
            self.creation_details[label]['created_count'] = len(self.creation_details[label]['resources'])

    def wait_for_create_resource(self, *, resource: dict):
        label = resource.get(Constants.LABEL)

        if label in self._added:
            self.logger.info(f"Waiting on Create: {label} using {self.label}: {self._added}")

            try:
                self.do_wait_for_create_resource(resource=resource)
            except (Exception, KeyboardInterrupt) as e:
                self.failed[label] = 'CREATE'
                failed_count = resource[Constants.RES_COUNT] - len(self.creation_details[label]['resources'])
                self.creation_details[label]['failed_count'] = failed_count
                raise e
            finally:
                self.creation_details[label]['created_count'] = len(self.creation_details[label]['resources'])

    def delete_resource(self, *, resource: dict):
        try:
            self.do_delete_resource(resource=resource)
        except (Exception, KeyboardInterrupt) as e:
            label = resource.get(Constants.LABEL)

            self.failed[label] = 'DELETE'
            raise e

    def get_state(self) -> ProviderState:
        def cleanup_attrs(attrs):
            attributes = attrs.copy()
            attributes.pop('logger', None)
            attributes.pop('label')
            attributes = {k: v for k, v in attributes.items() if not k.startswith('_')}
            return attributes

        services = [s for s in self.services if s.name in self.creation_details[s.label]["resources"]]
        service_states = [ServiceState(label=s.label, attributes=cleanup_attrs(vars(s))) for s in services]
        pending = [res['label'] for res in self.pending]
        pending_internal = [res['label'] for res in self.pending_internal]
        return ProviderState(self.label, dict(name=self.name), service_states,
                             pending, pending_internal, self.failed, self.creation_details)

    @abstractmethod
    def setup_environment(self):
        pass

    @abstractmethod
    def do_validate_resource(self, *, resource: dict):
        pass

    @abstractmethod
    def do_add_resource(self, *, resource: dict):
        pass

    @abstractmethod
    def do_create_resource(self, *, resource: dict):
        pass

    @abstractmethod
    def do_wait_for_create_resource(self, *, resource: dict):
        pass

    @abstractmethod
    def do_delete_resource(self, *, resource: dict):
        pass



import logging
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from .constants import Constants


def create_parser(usage='%(prog)s [options]',
                  description='',
                  formatter_class=None):
    formatter_class = formatter_class or RawDescriptionHelpFormatter
    return ArgumentParser(usage=usage, description=description, formatter_class=formatter_class)


def build_parser(*, manage_workflow, manage_sessions):
    description = (
        'Sense Workflow Tool'
        '\n'
        '\n'
        'Examples:'
        '\n'
        "      sense_workflow.py workflow --var-file vars.yml --session my_session -validate"
        '\n'
        "      sense_workflow.py workflow --config-dir . --session my_session -validate"
    )

    parser = create_parser(description=description)
    subparsers = parser.add_subparsers()
    workflow_parser = subparsers.add_parser('workflow', help='Manage sense workflows')
    workflow_parser.add_argument('-c', '--config-dir', type=str, default='.',
                                 help='config directory with .sense files. Defaults to current directory.',
                                 required=False)
    workflow_parser.add_argument('-v', '--var-file', type=str, default='',
                                 help="Yaml file with key-value pairs to override the variables' default values",
                                 required=False)
    workflow_parser.add_argument('-s', '--session', type=str, default='',
                                 help='friendly session name to help track a workflow', required=True)
    workflow_parser.add_argument('-p', '--policy-file', type=str, default='',
                                 help="Yaml stitching policy file",
                                 required=False)
    workflow_parser.add_argument('-validate', action='store_true', default=False,
                                 help='assembles and validates all .sense files  in the config directory')
    workflow_parser.add_argument('-plan', action='store_true', default=False, help='shows plan')
    workflow_parser.add_argument('-apply', action='store_true', default=False, help='create resources')
    workflow_parser.add_argument('-show', action='store_true', default=False, help='display resource.')
    workflow_parser.add_argument('-summary', action='store_true', default=False,
                                 help='display summary. used with -show')
    workflow_parser.add_argument('-json', action='store_true', default=False,
                                 help='use json output. relevant when used with -show or -plan')
    workflow_parser.add_argument('-destroy', action='store_true', default=False, help='delete resources')
    workflow_parser.set_defaults(dispatch_func=manage_workflow)

    sessions_parser = subparsers.add_parser('sessions', help='Manage sense sessions ')
    sessions_parser.add_argument('-show', action='store_true', default=False, help='display sessions')
    sessions_parser.add_argument('-json', action='store_true', default=False, help='use json format')
    sessions_parser.set_defaults(dispatch_func=manage_sessions)
    return parser


def get_log_level():
    import os
    return os.environ.get('SENSE_LOG_LEVEL', "INFO")


def get_log_location():
    import os
    return os.environ.get('SENSE_LOG_LOCATION', "/tmp/sense.log")


def get_formatter():
    fmt = "%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s"
    return logging.Formatter(fmt)


_LOGGER = None


def get_logger():
    global _LOGGER

    if not _LOGGER:
        _LOGGER = init_logger()

    return _LOGGER


def init_logger():
    from logging.handlers import RotatingFileHandler

    log_config = {'log-file': get_log_location(),
                  'log-level': get_log_level(),
                  'log-retain': 5,
                  'log-size': 5000000,
                  'logger': 'sense'}

    logger = logging.getLogger(str(log_config.get(Constants.PROPERTY_CONF_LOGGER, __name__)))
    log_level = log_config.get(Constants.PROPERTY_CONF_LOG_LEVEL, "INFO")
    logger.setLevel(log_level)

    formatter = get_formatter()
    file_handler = RotatingFileHandler(log_config.get(Constants.PROPERTY_CONF_LOG_FILE),
                                       backupCount=int(log_config.get(Constants.PROPERTY_CONF_LOG_RETAIN)),
                                       maxBytes=int(log_config.get(Constants.PROPERTY_CONF_LOG_SIZE)))

    file_handler.setFormatter(formatter)
    logger.propagate = False
    logger.addHandler(file_handler)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    global _LOGGER

    _LOGGER = logger
    return logger


def absolute_path(path):
    from pathlib import Path
    import os

    path = Path(path).expanduser().absolute()

    return os.path.realpath(str(path))


def load_as_ns_from_yaml(*, dir_path=None, content=None):
    import yaml
    import json
    from types import SimpleNamespace

    objs = []

    if dir_path:
        from pathlib import Path
        import os

        dir_path = Path(dir_path).expanduser().absolute()

        if not os.path.isdir(dir_path):
            raise Exception(f'Expected a directory {dir_path}')

        configs = [conf for conf in os.listdir(dir_path) if conf.endswith(Constants.SENSE_EXTENSION)]

        if not configs:
            raise Exception(f'No {Constants.SENSE_EXTENSION} config files found in  {dir_path}')

        for config in configs:
            file_name = os.path.join(dir_path, config)

            with open(file_name, 'r') as stream:
                obj = yaml.safe_load(stream)
                obj = json.loads(json.dumps(obj), object_hook=lambda dct: SimpleNamespace(**dct))
                objs.append(obj)
    else:
        obj = yaml.safe_load(content)
        obj = json.loads(json.dumps(obj), object_hook=lambda dct: SimpleNamespace(**dct))
        objs.append(obj)

    return objs


def load_yaml_from_file(file_name):
    import yaml
    from pathlib import Path

    path = Path(file_name).expanduser().absolute()

    with open(str(path), 'r') as stream:
        return yaml.safe_load(stream)


def load_vars(var_file):
    import yaml
    import os

    if not os.path.isfile(var_file):
        raise Exception(f'The supplied var-file {var_file} is invalid')

    with open(var_file, 'r') as stream:
        return yaml.safe_load(stream)


def get_base_dir(friendly_name):
    from pathlib import Path
    import os

    base_dir = os.path.join(str(Path.home()), '.sense', 'sessions', friendly_name)
    os.makedirs(base_dir, exist_ok=True)
    return base_dir


def dump_sessions(to_json: bool):
    from .state_utils import load_meta_data
    from pathlib import Path
    import os
    import sys

    base_dir = os.path.join(str(Path.home()), '.sense', 'sessions')
    os.makedirs(base_dir, exist_ok=True)
    sessions = os.listdir(base_dir)
    sessions = [dict(session=s,
                     config_dir=load_meta_data(s, 'config_dir')) for s in sessions]

    if to_json:
        import json

        sys.stdout.write(json.dumps(sessions, default=lambda o: o.__dict__, indent=3))
    else:
        import yaml

        sys.stdout.write(yaml.dump(sessions))

    return sessions


def get_counters(*, states):
    services = pending = failed = 0

    for state in states:
        pending += len(state.pending)
        pending += len(state.pending_internal)
        services += len(state.service_states)

        for key, detail in state.creation_details.items():
            failed += detail['failed_count']

    return services, pending, failed


# noinspection PyBroadException
def can_read(path: str):
    from pathlib import Path

    path = str(Path(path).expanduser().absolute())

    try:
        with open(path, 'r') as f:
            f.read()
        return True
    except Exception:
        return False


# noinspection PyBroadException
def can_read_json(path: str):
    try:
        with open(path, 'r') as fp:
            import json

            json.load(fp)

        return True
    except Exception:
        return False




from sense.workflow.base.constants import Constants
from sense.workflow.base.exceptions import ResourceTypeNotSupported, ProviderException
from sense.workflow.base.utils import get_logger
from sense.workflow.provider.provider import Provider
from .sense_constants import *
from .sense_exceptions import SenseException

logger = get_logger()


class SenseProvider(Provider):
    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        self.supported_resources = [Constants.RES_TYPE_SERVICE]
        self._client = None
        self.service_handler = None
        self.address_handler = None

    def setup_environment(self):
        for attr in SENSE_CONF_ATTRS:
            if self.config.get(attr.lower()) is None and self.config.get(attr.upper()) is None:
                raise ProviderException(f"{self.name}: Expecting a value for {attr}")

    def _init_client(self):
        if not self.client:
            self.logger.debug(f"{self.name}: Initializing sense client {self.label}")
            from sense.client.requestwrapper import RequestWrapper

            self._client = RequestWrapper(self.config)
            self.logger.info(f"{self.name}: Initialized sense client {self.label}")

    @property
    def client(self):
        return self._client

    def supports_modify(self):
        return True

    def do_validate_resource(self, *, resource: dict):
        label = resource[Constants.LABEL]
        rtype = resource[Constants.RES_TYPE]

        if rtype not in self.supported_resources:
            raise ResourceTypeNotSupported(f"{rtype} for {label}")

        if Constants.PROFILE not in resource and Constants.POOL not in resource:
            raise SenseException(f"Must have a profile or pool for {label}")

    def _get_handler(self, resource):
        def handler_from_dict(adict: dict):
            from .service_handler import ServiceHandler
            from .address_handler import AddressHandler

            if Constants.PROFILE in adict:
                self.service_handler = self.service_handler or ServiceHandler(provider=self)
                return self.service_handler
            elif Constants.POOL in adict:
                self.address_handler = self.address_handler or AddressHandler(provider=self)
                return self.address_handler
            else:
                return None

        handler = handler_from_dict(resource)

        if handler is None:
            if resource[Constants.SAVED_STATES]:
                handler = handler_from_dict(resource[Constants.SAVED_STATES][0].attributes)

        if handler is None:
            raise SenseException(f'unkown resource {resource}')

        return handler

    def do_add_resource(self, *, resource: dict):
        assert resource[Constants.RES_TYPE] in self.supported_resources
        self._init_client()

        handler = self._get_handler(resource)
        handler.add_resource(resource=resource)

    def do_create_resource(self, *, resource: dict):
        assert resource[Constants.RES_TYPE] in self.supported_resources
        self._init_client()
        handler = self._get_handler(resource)
        handler.create_resource(resource=resource)

    def do_wait_for_create_resource(self, *, resource: dict):
        assert resource[Constants.RES_TYPE] in self.supported_resources
        self._init_client()
        handler = self._get_handler(resource)
        handler.wait_for_create_resource(resource=resource)

    def do_delete_resource(self, *, resource: dict):
        assert resource[Constants.RES_TYPE] in self.supported_resources
        self._init_client()
        handler = self._get_handler(resource)
        handler.delete_resource(resource=resource)



from abc import ABC
from collections import namedtuple


class Resource(ABC):
    def __init__(self, label, name: str):
        self.label = label
        self.name = name

    def get_label(self) -> str:
        return self.label

    def get_name(self) -> str:
        return self.name


class Service(Resource):
    def __init__(self, *, label, name: str):
        super().__init__(label, name)


ResolvedDependency = namedtuple("ResolvedDependency", "resource_label attr value")



#!/usr/bin/env python3
"""API Client for SENSE-0 get Token and Config"""
import os
import json
import requests
from sense.common import classwrapper
from sense.common import getConfig

requests.packages.urllib3.disable_warnings()

@classwrapper
class ApiClient():
    """API Client for SENSE-0 get Token and Config"""
    def __init__(self, config):
        # For now only pass config file; Later all params
        self.config = config or getConfig()
        self._validateConfig()
        self._setDefaults()
        self.token = None
        self._getToken()

    def _getToken(self):
        """Get Token from SENSE-0 Auth API"""
        data = {'grant_type': 'password',
                'username': self.config['USERNAME'],
                'password': self.config['PASSWORD']}
        tokenResponse = requests.post(self.config['AUTH_ENDPOINT'],
                                      data=data,
                                      verify=self.config['verify'],
                                      allow_redirects=self.config['allow_redirects'],
                                      auth=(self.config['CLIENT_ID'], self.config['SECRET']),
                                      timeout=int(os.environ.get('SENSE_TIMEOUT', 60)))
        self.token = json.loads(tokenResponse.text)
        if 'error' in self.token.keys() and 'error_description' in self.token.keys():
            raise Exception(f"Failed to get token. Bad credentials? Error: {self.token['error_description']}")
        self._setHeaders()

    def _setHeaders(self, content='json', accept='json'):
        """Set Headers for API calls"""
        self.config['headers'] = {'Content-type': f'application/{content}', 'Accept': f'application/{accept}',
                                  'Authorization': 'Bearer ' + self.token['access_token']}

    def _refreshToken(self):
        """Refresh Token from SENSE-0 Auth API"""
        self._getToken()

    def _setDefaults(self):
        """Set Defaults for Config"""
        for key, val in {'verify': False,
                         'allow_redirects': False,
                         'REST_API': self.config['API_ENDPOINT']}.items():
            if key not in self.config.keys():
                self.config[key] = val

    def _validateConfig(self):
        """Validate Config"""
        for param in ['AUTH_ENDPOINT', 'API_ENDPOINT', 'USERNAME', 'PASSWORD', 'CLIENT_ID', 'SECRET']:
            if param not in self.config.keys():
                raise Exception(f"Config parameter {param} is not set")



import json
import logging
from types import SimpleNamespace

from sense.workflow.base.constants import Constants
from sense.workflow.base.resource_models import ResolvedDependency, Resource


def cleanup_attrs(attrs):
    attributes = attrs.copy()
    attributes.pop('logger', None)
    attributes.pop('label')
    attributes = {k: v for k, v in attributes.items() if not k.startswith('_')}
    return attributes


class DependencyResolver:
    def __init__(self, *, label, logger: logging.Logger, external=True):
        self.label = label
        self.logger = logger

        if external:
            self.dependency_label = Constants.EXTERNAL_DEPENDENCIES
            self.resolved_dependency_label = Constants.RESOLVED_EXTERNAL_DEPENDENCIES
        else:
            self.dependency_label = Constants.INTERNAL_DEPENDENCIES
            self.resolved_dependency_label = Constants.RESOLVED_INTERNAL_DEPENDENCIES

    def check_if_dependencies_are_resolved(self, *, resource: dict):
        label = resource[Constants.LABEL]
        self.logger.info(f"Checking if all dependencies are resolved for {label} using {self.label}")

        if len(resource[self.dependency_label]) == len(resource[self.resolved_dependency_label]):
            ok = True

            for dependency in resource[self.dependency_label]:
                count = dependency.resource.attributes.get(Constants.RES_COUNT, 1)
                assert count > 0

                temp = resource[self.resolved_dependency_label]
                criteria = (dependency.key, dependency.resource.label)
                resolved_dependencies = [rd for rd in temp if (rd.attr, rd.resource_label) == criteria]

                assert len(resolved_dependencies) == 1
                resolved_dependency = resolved_dependencies[0]

                if len(resolved_dependency.value) != count:
                    ok = False
                    break

            self.logger.info(f"Checking if all dependencies are resolved for {label} using {self.label}:ret={ok}")
            return ok

        self.logger.info(f"Checking if all dependencies are resolved for {label} using {self.label}:ret=false")
        return False

    def resolve_dependency(self, *, resource: dict, from_resource: Resource):
        from_resource_dict = vars(from_resource)
        label = resource[Constants.LABEL]

        for dependency in resource[self.dependency_label]:
            count = dependency.resource.attributes.get(Constants.RES_COUNT, 1)
            assert count > 0

            if dependency.resource.label == from_resource.label:
                try:
                    if not dependency.attribute:
                        value = from_resource
                    else:
                        attrs = json.loads(json.dumps(cleanup_attrs(from_resource_dict)),
                                           object_hook=lambda dct: SimpleNamespace(**dct))
                        value = eval("attrs." + dependency.attribute, {}, {'attrs': locals()['attrs']})

                    if value and isinstance(value, list):
                        value = tuple(value)

                    self.logger.info(
                        f"Resolving: {dependency} for {label}: value={value} using {self.label}")

                    if value:
                        resolved_dependencies = resource[self.resolved_dependency_label]
                        from typing import List, Dict, Union
                        found: Union[ResolvedDependency, None] = None

                        for resolved_dependency in resolved_dependencies:
                            if resolved_dependency.attr == dependency.key \
                                    and resolved_dependency.resource_label == dependency.resource.label:
                                found = resolved_dependency
                                break

                        if not found:
                            resolved_dependency = ResolvedDependency(resource_label=dependency.resource.label,
                                                                     attr=dependency.key,
                                                                     value=(value,))
                            resolved_dependencies.append(resolved_dependency)
                            self.logger.info(f"Resolved dependency {dependency} for {label} using {self.label}")
                        elif len(found.value) < count:
                            resolved_dependency = ResolvedDependency(resource_label=dependency.resource.label,
                                                                     attr=dependency.key,
                                                                     value=(value,) + found.value)
                            resolved_dependencies.remove(found)
                            resolved_dependencies.append(resolved_dependency)
                            self.logger.info(f"Resolved dependency {dependency} for {label} using {self.label}")
                    else:
                        self.logger.warning(
                            f"Could not resolve {dependency} for {label} using {self.label}")
                except Exception as e:
                    self.logger.warning(
                        f"Severe Exception occurred while resolving dependency: {e} using {self.label}")
                    self.logger.error(e, exc_info=True)

    def extract_values(self, *, resource: dict):
        assert resource[self.dependency_label]

        for dependency in resource[self.dependency_label]:
            attribute = dependency.key

            label = resource.get(Constants.LABEL)
            self.logger.debug(f"Extracting Values: {label}:{attribute} using {self.label}")

            resolved_dependencies = [rd for rd in resource[self.resolved_dependency_label]
                                     if rd.attr == attribute]

            assert resolved_dependencies

            values = [rd.value for rd in resolved_dependencies]
            resource[attribute] = [rd.value for rd in resolved_dependencies]
            self.logger.info(f"Extracted Values: {values}:{label}:{attribute} using {self.label}")



# coding: utf-8

"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class Terminal(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'uri': 'str',
        'vlan_tag': 'str',
        'ipv4_prefix_list': 'str',
        'assign_ip': 'bool'
    }

    attribute_map = {
        'uri': 'uri',
        'vlan_tag': 'vlan_tag',
        'ipv4_prefix_list': 'ipv4_prefix_list',
        'assign_ip': 'assign_ip'
    }

    def __init__(self, uri=None, vlan_tag=None, ipv4_prefix_list=None, assign_ip=None):  # noqa: E501
        """Terminal - a model defined in Swagger"""  # noqa: E501
        self._uri = None
        self._vlan_tag = None
        self._ipv4_prefix_list = None
        self._assign_ip = None
        self.discriminator = None
        if uri is not None:
            self.uri = uri
        if vlan_tag is not None:
            self.vlan_tag = vlan_tag
        if ipv4_prefix_list is not None:
            self.ipv4_prefix_list = ipv4_prefix_list
        if assign_ip is not None:
            self.assign_ip = assign_ip

    @property
    def uri(self):
        """Gets the uri of this Terminal.  # noqa: E501


        :return: The uri of this Terminal.  # noqa: E501
        :rtype: str
        """
        return self._uri

    @uri.setter
    def uri(self, uri):
        """Sets the uri of this Terminal.


        :param uri: The uri of this Terminal.  # noqa: E501
        :type: str
        """

        self._uri = uri

    @property
    def vlan_tag(self):
        """Gets the vlan_tag of this Terminal.  # noqa: E501


        :return: The vlan_tag of this Terminal.  # noqa: E501
        :rtype: str
        """
        return self._vlan_tag

    @vlan_tag.setter
    def vlan_tag(self, vlan_tag):
        """Sets the vlan_tag of this Terminal.


        :param vlan_tag: The vlan_tag of this Terminal.  # noqa: E501
        :type: str
        """

        self._vlan_tag = vlan_tag

    @property
    def ipv4_prefix_list(self):
        """Gets the ipv4_prefix_list of this Terminal.  # noqa: E501


        :return: The ipv4_prefix_list of this Terminal.  # noqa: E501
        :rtype: str
        """
        return self._ipv4_prefix_list

    @ipv4_prefix_list.setter
    def ipv4_prefix_list(self, ipv4_prefix_list):
        """Sets the ipv4_prefix_list of this Terminal.


        :param ipv4_prefix_list: The ipv4_prefix_list of this Terminal.  # noqa: E501
        :type: str
        """

        self._ipv4_prefix_list = ipv4_prefix_list

    @property
    def assign_ip(self):
        """Gets the assign_ip of this Terminal.  # noqa: E501


        :return: The assign_ip of this Terminal.  # noqa: E501
        :rtype: bool
        """
        return self._assign_ip

    @assign_ip.setter
    def assign_ip(self, assign_ip):
        """Sets the assign_ip of this Terminal.


        :param assign_ip: The assign_ip of this Terminal.  # noqa: E501
        :type: bool
        """

        self._assign_ip = assign_ip

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Terminal, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Terminal):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class LoggingConfiguration(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'loggers': 'LoggingConfigurationLoggers',
        'filter': 'str',
        'main_rows': 'float',
        'main_usage': 'float',
        'archived_rows': 'float',
        'archived_usage': 'float'
    }

    attribute_map = {
        'loggers': 'loggers',
        'filter': 'filter',
        'main_rows': 'mainRows',
        'main_usage': 'mainUsage',
        'archived_rows': 'archivedRows',
        'archived_usage': 'archivedUsage'
    }

    def __init__(self, loggers=None, filter=None, main_rows=None, main_usage=None, archived_rows=None, archived_usage=None):  # noqa: E501
        """LoggingConfiguration - a model defined in Swagger"""  # noqa: E501
        self._loggers = None
        self._filter = None
        self._main_rows = None
        self._main_usage = None
        self._archived_rows = None
        self._archived_usage = None
        self.discriminator = None
        self.loggers = loggers
        if filter is not None:
            self.filter = filter
        if main_rows is not None:
            self.main_rows = main_rows
        if main_usage is not None:
            self.main_usage = main_usage
        if archived_rows is not None:
            self.archived_rows = archived_rows
        if archived_usage is not None:
            self.archived_usage = archived_usage

    @property
    def loggers(self):
        """Gets the loggers of this LoggingConfiguration.  # noqa: E501


        :return: The loggers of this LoggingConfiguration.  # noqa: E501
        :rtype: LoggingConfigurationLoggers
        """
        return self._loggers

    @loggers.setter
    def loggers(self, loggers):
        """Sets the loggers of this LoggingConfiguration.


        :param loggers: The loggers of this LoggingConfiguration.  # noqa: E501
        :type: LoggingConfigurationLoggers
        """
        if loggers is None:
            raise ValueError("Invalid value for `loggers`, must not be `None`")  # noqa: E501

        self._loggers = loggers

    @property
    def filter(self):
        """Gets the filter of this LoggingConfiguration.  # noqa: E501

        Filter configuration string, for dropping matching messages from logging.  # noqa: E501

        :return: The filter of this LoggingConfiguration.  # noqa: E501
        :rtype: str
        """
        return self._filter

    @filter.setter
    def filter(self, filter):
        """Sets the filter of this LoggingConfiguration.

        Filter configuration string, for dropping matching messages from logging.  # noqa: E501

        :param filter: The filter of this LoggingConfiguration.  # noqa: E501
        :type: str
        """

        self._filter = filter

    @property
    def main_rows(self):
        """Gets the main_rows of this LoggingConfiguration.  # noqa: E501

        Number of rows present in the main logging table.  # noqa: E501

        :return: The main_rows of this LoggingConfiguration.  # noqa: E501
        :rtype: float
        """
        return self._main_rows

    @main_rows.setter
    def main_rows(self, main_rows):
        """Sets the main_rows of this LoggingConfiguration.

        Number of rows present in the main logging table.  # noqa: E501

        :param main_rows: The main_rows of this LoggingConfiguration.  # noqa: E501
        :type: float
        """

        self._main_rows = main_rows

    @property
    def main_usage(self):
        """Gets the main_usage of this LoggingConfiguration.  # noqa: E501

        Bytes usage in the main logging table.  # noqa: E501

        :return: The main_usage of this LoggingConfiguration.  # noqa: E501
        :rtype: float
        """
        return self._main_usage

    @main_usage.setter
    def main_usage(self, main_usage):
        """Sets the main_usage of this LoggingConfiguration.

        Bytes usage in the main logging table.  # noqa: E501

        :param main_usage: The main_usage of this LoggingConfiguration.  # noqa: E501
        :type: float
        """

        self._main_usage = main_usage

    @property
    def archived_rows(self):
        """Gets the archived_rows of this LoggingConfiguration.  # noqa: E501

        Number of rows present in the archived logging table.  # noqa: E501

        :return: The archived_rows of this LoggingConfiguration.  # noqa: E501
        :rtype: float
        """
        return self._archived_rows

    @archived_rows.setter
    def archived_rows(self, archived_rows):
        """Sets the archived_rows of this LoggingConfiguration.

        Number of rows present in the archived logging table.  # noqa: E501

        :param archived_rows: The archived_rows of this LoggingConfiguration.  # noqa: E501
        :type: float
        """

        self._archived_rows = archived_rows

    @property
    def archived_usage(self):
        """Gets the archived_usage of this LoggingConfiguration.  # noqa: E501

        Bytes usage in the archived logging table.  # noqa: E501

        :return: The archived_usage of this LoggingConfiguration.  # noqa: E501
        :rtype: float
        """
        return self._archived_usage

    @archived_usage.setter
    def archived_usage(self, archived_usage):
        """Sets the archived_usage of this LoggingConfiguration.

        Bytes usage in the archived logging table.  # noqa: E501

        :param archived_usage: The archived_usage of this LoggingConfiguration.  # noqa: E501
        :type: float
        """

        self._archived_usage = archived_usage

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(LoggingConfiguration, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LoggingConfiguration):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class VerificationResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'delta_uuid': 'str',
        'creation_time': 'str',
        'verified_reduction': 'str',
        'verified_addition': 'str',
        'unverified_reduction': 'str',
        'unverified_addition': 'str',
        'reduction': 'str',
        'addition': 'str'
    }

    attribute_map = {
        'delta_uuid': 'delta_uuid',
        'creation_time': 'creation_time',
        'verified_reduction': 'verified_reduction',
        'verified_addition': 'verified_addition',
        'unverified_reduction': 'unverified_reduction',
        'unverified_addition': 'unverified_addition',
        'reduction': 'reduction',
        'addition': 'addition'
    }

    def __init__(self, delta_uuid=None, creation_time=None, verified_reduction=None, verified_addition=None, unverified_reduction=None, unverified_addition=None, reduction=None, addition=None):  # noqa: E501
        """VerificationResult - a model defined in Swagger"""  # noqa: E501
        self._delta_uuid = None
        self._creation_time = None
        self._verified_reduction = None
        self._verified_addition = None
        self._unverified_reduction = None
        self._unverified_addition = None
        self._reduction = None
        self._addition = None
        self.discriminator = None
        self.delta_uuid = delta_uuid
        self.creation_time = creation_time
        if verified_reduction is not None:
            self.verified_reduction = verified_reduction
        if verified_addition is not None:
            self.verified_addition = verified_addition
        if unverified_reduction is not None:
            self.unverified_reduction = unverified_reduction
        if unverified_addition is not None:
            self.unverified_addition = unverified_addition
        self.reduction = reduction
        self.addition = addition

    @property
    def delta_uuid(self):
        """Gets the delta_uuid of this VerificationResult.  # noqa: E501

        UUID of delta being verified.  # noqa: E501

        :return: The delta_uuid of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._delta_uuid

    @delta_uuid.setter
    def delta_uuid(self, delta_uuid):
        """Sets the delta_uuid of this VerificationResult.

        UUID of delta being verified.  # noqa: E501

        :param delta_uuid: The delta_uuid of this VerificationResult.  # noqa: E501
        :type: str
        """
        if delta_uuid is None:
            raise ValueError("Invalid value for `delta_uuid`, must not be `None`")  # noqa: E501

        self._delta_uuid = delta_uuid

    @property
    def creation_time(self):
        """Gets the creation_time of this VerificationResult.  # noqa: E501

        When this verification run started.  # noqa: E501

        :return: The creation_time of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._creation_time

    @creation_time.setter
    def creation_time(self, creation_time):
        """Sets the creation_time of this VerificationResult.

        When this verification run started.  # noqa: E501

        :param creation_time: The creation_time of this VerificationResult.  # noqa: E501
        :type: str
        """
        if creation_time is None:
            raise ValueError("Invalid value for `creation_time`, must not be `None`")  # noqa: E501

        self._creation_time = creation_time

    @property
    def verified_reduction(self):
        """Gets the verified_reduction of this VerificationResult.  # noqa: E501

        JSON string of the verified reduction model.  # noqa: E501

        :return: The verified_reduction of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._verified_reduction

    @verified_reduction.setter
    def verified_reduction(self, verified_reduction):
        """Sets the verified_reduction of this VerificationResult.

        JSON string of the verified reduction model.  # noqa: E501

        :param verified_reduction: The verified_reduction of this VerificationResult.  # noqa: E501
        :type: str
        """

        self._verified_reduction = verified_reduction

    @property
    def verified_addition(self):
        """Gets the verified_addition of this VerificationResult.  # noqa: E501

        JSON string of the verified addition model.  # noqa: E501

        :return: The verified_addition of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._verified_addition

    @verified_addition.setter
    def verified_addition(self, verified_addition):
        """Sets the verified_addition of this VerificationResult.

        JSON string of the verified addition model.  # noqa: E501

        :param verified_addition: The verified_addition of this VerificationResult.  # noqa: E501
        :type: str
        """

        self._verified_addition = verified_addition

    @property
    def unverified_reduction(self):
        """Gets the unverified_reduction of this VerificationResult.  # noqa: E501

        JSON string of the unverified reduction model.  # noqa: E501

        :return: The unverified_reduction of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._unverified_reduction

    @unverified_reduction.setter
    def unverified_reduction(self, unverified_reduction):
        """Sets the unverified_reduction of this VerificationResult.

        JSON string of the unverified reduction model.  # noqa: E501

        :param unverified_reduction: The unverified_reduction of this VerificationResult.  # noqa: E501
        :type: str
        """

        self._unverified_reduction = unverified_reduction

    @property
    def unverified_addition(self):
        """Gets the unverified_addition of this VerificationResult.  # noqa: E501

        JSON string of the unverified addition model.  # noqa: E501

        :return: The unverified_addition of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._unverified_addition

    @unverified_addition.setter
    def unverified_addition(self, unverified_addition):
        """Sets the unverified_addition of this VerificationResult.

        JSON string of the unverified addition model.  # noqa: E501

        :param unverified_addition: The unverified_addition of this VerificationResult.  # noqa: E501
        :type: str
        """

        self._unverified_addition = unverified_addition

    @property
    def reduction(self):
        """Gets the reduction of this VerificationResult.  # noqa: E501

        String boolean of whether the reduction is present.  # noqa: E501

        :return: The reduction of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._reduction

    @reduction.setter
    def reduction(self, reduction):
        """Sets the reduction of this VerificationResult.

        String boolean of whether the reduction is present.  # noqa: E501

        :param reduction: The reduction of this VerificationResult.  # noqa: E501
        :type: str
        """
        if reduction is None:
            raise ValueError("Invalid value for `reduction`, must not be `None`")  # noqa: E501
        allowed_values = ["true", "false"]  # noqa: E501
        if reduction not in allowed_values:
            raise ValueError(
                "Invalid value for `reduction` ({0}), must be one of {1}"  # noqa: E501
                .format(reduction, allowed_values)
            )

        self._reduction = reduction

    @property
    def addition(self):
        """Gets the addition of this VerificationResult.  # noqa: E501

        String boolean of whether the addition is present.  # noqa: E501

        :return: The addition of this VerificationResult.  # noqa: E501
        :rtype: str
        """
        return self._addition

    @addition.setter
    def addition(self, addition):
        """Sets the addition of this VerificationResult.

        String boolean of whether the addition is present.  # noqa: E501

        :param addition: The addition of this VerificationResult.  # noqa: E501
        :type: str
        """
        if addition is None:
            raise ValueError("Invalid value for `addition`, must not be `None`")  # noqa: E501
        allowed_values = ["true", "false"]  # noqa: E501
        if addition not in allowed_values:
            raise ValueError(
                "Invalid value for `addition` ({0}), must be one of {1}"  # noqa: E501
                .format(addition, allowed_values)
            )

        self._addition = addition

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(VerificationResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, VerificationResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import json
import os
import sys
from typing import List, Dict

import yaml

from .exceptions import StateException
from .state_models import ProviderState
from .state_models import get_dumper
from .utils import get_base_dir


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        else:
            return obj.__dict__


def dump_plan(*, resources, to_json: bool, summary: bool = False):
    from collections import namedtuple
    from .constants import Constants
    import sys

    plan = {}
    ResourceSummary = namedtuple("ResourceSummary", "label attributes")

    if not summary:
        summaries = []

        ResourceDetails = namedtuple("ResourceDetails", "label attributes")

        for resource in resources:
            resource_dict = resource.attributes.copy()
            del_attrs = [Constants.EXTERNAL_DEPENDENCIES,
                         Constants.RESOLVED_EXTERNAL_DEPENDENCIES,
                         Constants.INTERNAL_DEPENDENCIES,
                         Constants.RESOLVED_INTERNAL_DEPENDENCIES,
                         Constants.SAVED_STATES]

            for attr in del_attrs:
                del resource_dict[attr]

            stringify_attrs = [Constants.PROVIDER]

            for attr in stringify_attrs:
                if attr in resource_dict:
                    resource_dict[attr] = str(resource_dict[attr])

            label = resource_dict.pop(Constants.LABEL)
            summaries.append(ResourceDetails(label=label, attributes=resource_dict))
            plan['resource_details'] = summaries

        summaries = []

        for resource in resources:
            resource_dict = {}

            label = resource.attributes[Constants.LABEL]

            import copy

            details = copy.deepcopy(resource.attributes[Constants.RES_CREATION_DETAILS])
            resource_dict[Constants.RES_CREATION_DETAILS] = details
            summaries.append(ResourceSummary(label=label, attributes=resource_dict))

        plan['resource_creation_details'] = summaries

    summaries = []
    to_be_created = 0
    to_be_deleted = 0
    provider_resource_map = {}

    for resource in resources:
        details = resource.attributes[Constants.RES_CREATION_DETAILS]
        provider_supports_modifiable = details['provider_supports_modifiable']

        if not provider_supports_modifiable:
            provider_resource_map[resource.provider.label] = False

    for resource in resources:
        details = resource.attributes[Constants.RES_CREATION_DETAILS]
        provider_supports_modifiable = details['provider_supports_modifiable']

        if not provider_supports_modifiable:
            in_config_file = details['in_config_file']
            changed = not in_config_file or details['total_count'] != details['created_count']
            provider_resource_map[resource.provider.label] = provider_resource_map[resource.provider.label] or changed

    for resource in resources:
        resource_dict = {}
        label = resource.attributes[Constants.LABEL]
        details = resource.attributes[Constants.RES_CREATION_DETAILS]
        provider_supports_modifiable = details['provider_supports_modifiable']
        in_config_file = details['in_config_file']

        if provider_supports_modifiable:
            if in_config_file:
                if details['total_count'] > details['created_count']:
                    resource_dict['to_be_created'] = details['total_count'] - details['created_count']
                    resource_dict['to_be_deleted'] = 0
                else:
                    resource_dict['to_be_created'] = 0
                    resource_dict['to_be_deleted'] = details['created_count'] - details['total_count']
            else:
                resource_dict['to_be_created'] = 0
                resource_dict['to_be_deleted'] = details['created_count']
        elif provider_resource_map[resource.provider.label]:
            if in_config_file:
                resource_dict['to_be_created'] = details['total_count']
                resource_dict['to_be_deleted'] = details['created_count']
            else:
                resource_dict['to_be_created'] = 0
                resource_dict['to_be_deleted'] = details['created_count']
        else:
            resource_dict['to_be_created'] = 0
            resource_dict['to_be_deleted'] = 0

        to_be_created += resource_dict['to_be_created']
        to_be_deleted += resource_dict['to_be_deleted']
        summaries.append(ResourceSummary(label=label, attributes=resource_dict))

    plan['summaries'] = summaries

    if to_json:
        sys.stdout.write(json.dumps(plan, cls=SetEncoder, indent=3))
    else:
        sys.stdout.write(yaml.dump(plan, Dumper=get_dumper(), default_flow_style=False, sort_keys=False))

    return to_be_created, to_be_deleted


def dump_resources(*, resources, to_json: bool, summary: bool = False):
    if summary:
        summaries = []
        from collections import namedtuple
        from .constants import Constants

        ResourceSummary = namedtuple("ResourceSummary", "label attributes")

        for resource in resources:
            resource_dict = resource.attributes.copy()
            del_attrs = [Constants.EXTERNAL_DEPENDENCIES,
                         Constants.RESOLVED_EXTERNAL_DEPENDENCIES,
                         Constants.INTERNAL_DEPENDENCIES,
                         Constants.RESOLVED_INTERNAL_DEPENDENCIES,
                         Constants.SAVED_STATES]

            for attr in del_attrs:
                del resource_dict[attr]

            stringify_attrs = [Constants.PROVIDER]

            for attr in stringify_attrs:
                if attr in resource_dict:
                    resource_dict[attr] = str(resource_dict[attr])

            label = resource_dict.pop(Constants.LABEL)
            summaries.append(ResourceSummary(label=label, attributes=resource_dict))

        resources = summaries

    if to_json:
        sys.stdout.write(json.dumps(resources, cls=SetEncoder, indent=3))
    else:
        sys.stdout.write(yaml.dump(resources, Dumper=get_dumper(), default_flow_style=False, sort_keys=False))


def dump_objects(objects, to_json: bool):
    if to_json:
        sys.stdout.write(json.dumps(objects, cls=SetEncoder, indent=3))
    else:
        sys.stdout.write(yaml.dump(objects, Dumper=get_dumper(), default_flow_style=False, sort_keys=False))


def dump_states(states, to_json: bool, summary: bool = False):
    temp = []

    if not summary:
        for provider_state in states:
            for resource_state in provider_state.states():
                temp.append(resource_state)

    if summary:
        for provider_state in states:
            for service_state in provider_state.service_states:
                attributes = dict()
                props = ['name', 'id', 'state', 'profile', 'manifest', 'address', 'hosts']

                for prop in props:
                    if prop in service_state.attributes:
                        attributes[prop] = service_state.attributes[prop]

                service_state.attributes = attributes
                temp.append(service_state)

    states = temp
    output = dict()
    services = []

    for state in states:
        assert state.is_service_state
        services.append(state)

    output['services'] = services

    if to_json:
        sys.stdout.write(json.dumps(output, cls=SetEncoder, indent=3))
    else:
        sys.stdout.write(
            yaml.dump(output,
                      Dumper=get_dumper(),
                      width=float("inf"), default_flow_style=False, sort_keys=False))


def load_meta_data(friendly_name: str, attr=None):
    file_path = os.path.join(get_base_dir(friendly_name), friendly_name + '_meta.yml')

    if os.path.exists(file_path):
        with open(file_path, 'r') as stream:
            try:
                ret = yaml.safe_load(stream)

                if ret is not None:
                    return ret.get(attr) if attr else ret
            except Exception as e:
                raise StateException(f'Exception while loading state at {file_path}:{e}')

    return None if attr else dict()


def load_states(friendly_name) -> List[ProviderState]:
    import yaml
    import os

    file_path = os.path.join(get_base_dir(friendly_name), friendly_name + '.yml')

    if os.path.exists(file_path):
        with open(file_path, 'r') as stream:
            try:
                from .state_models import init_loader
                init_loader()
                ret = yaml.safe_load(stream)

                if ret is not None:
                    return ret
            except Exception as e:
                from .exceptions import StateException

                raise StateException(f'Exception while loading state at {file_path}:{e}')

    return list()


def load_states_as_dict(friendly_name) -> Dict[str, ProviderState]:
    states = load_states(friendly_name)
    state_map = {}

    for state in states:
        state_map[state.label] = state

    return state_map


def save_meta_data(meta_data: dict, friendly_name: str):
    import yaml
    import os

    file_path = os.path.join(get_base_dir(friendly_name), friendly_name + '_meta.yml')

    with open(file_path, "w") as stream:
        try:
            stream.write(yaml.dump(meta_data, Dumper=get_dumper()))
        except Exception as e:
            raise StateException(f'Exception while saving state at temp file {file_path}:{e}')


def save_states(states: List[ProviderState], friendly_name: str):
    import yaml
    import os

    file_path = os.path.join(get_base_dir(friendly_name), friendly_name + '.yml')
    temp_file_path = file_path + ".temp"

    with open(temp_file_path, "w") as stream:
        try:
            stream.write(yaml.dump(states, Dumper=get_dumper(), default_flow_style=False, sort_keys=False))
        except Exception as e:
            raise StateException(f'Exception while saving state at temp file {temp_file_path}:{e}')

    import shutil

    shutil.move(temp_file_path, file_path)


def reconcile_state(provider_state: ProviderState, saved_provider_state: ProviderState):
    assert provider_state.number_of_created_resources() != provider_state.number_of_total_resources()

    creation_details = provider_state.creation_details

    for resource_state in saved_provider_state.states():
        resource_label = resource_state.label

        if resource_label in creation_details:
            resource_details = creation_details[resource_label]

            if resource_details['total_count'] == resource_details['created_count']:
                continue

        provider_state.add_if_not_found(resource_state)


def reconcile_states(provider_states: List[ProviderState], friendly_name: str) -> List[ProviderState]:
    saved_states_map = load_states_as_dict(friendly_name)
    if next(filter(lambda s: s.number_of_created_resources() > 0, saved_states_map.values()), None) is None:
        return provider_states

    reconciled_states = []

    for provider_state in provider_states:
        if provider_state.label not in saved_states_map:
            reconciled_states.append(provider_state)
        elif provider_state.number_of_created_resources() == provider_state.number_of_total_resources():
            reconciled_states.append(provider_state)
        else:
            saved_provider_state = saved_states_map.get(provider_state.label)
            reconcile_state(provider_state=provider_state, saved_provider_state=saved_provider_state)
            reconciled_states.append(provider_state)

    return reconciled_states


def load_sessions():
    from pathlib import Path

    base_dir = os.path.join(str(Path.home()), '.sense', 'sessions')
    os.makedirs(base_dir, exist_ok=True)
    sessions = os.listdir(base_dir)
    return sessions


def destroy_session(friendly_name: str):
    import shutil

    dir_path = get_base_dir(friendly_name)
    shutil.rmtree(dir_path)



#!/usr/bin/env python3
# coding: utf-8
from sense.client.requestwrapper import RequestWrapper
from sense.common import classwrapper


@classwrapper
class TaskApi():

    def __init__(self, req_wrapper=None):
        if req_wrapper is None:
            self.client = RequestWrapper()
        else:
            self.client = req_wrapper
        if 'SI_UUID' in self.client.config:
            self.si_uuid = self.client.config['SI_UUID']
        else:
            self.si_uuid = None

    def get_tasks(self, **kwargs):
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_tasks_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.get_tasks_with_http_info(**kwargs)  # noqa: E501
            return data

    def get_tasks_with_http_info(self, **kwargs):
        all_params = ['async_req', '_return_http_data_only', '_preload_content', '_request_timeout', 'assigned']
        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method task_query" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('GET', f'/task/assigned/{kwargs["assigned"]}')

    def get_tasks_agent_status(self, **kwargs):
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_tasks_agent_status_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.get_tasks_agent_status_with_http_info(**kwargs)  # noqa: E501
            return data

    def get_tasks_agent_status_with_http_info(self, **kwargs):
        all_params = ['async_req', '_return_http_data_only', '_preload_content', '_request_timeout', 'assigned', 'status']
        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method task_query" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('GET', f'/task/assigned/{kwargs["assigned"]}/status/{kwargs["status"]}')

    def get_task(self, **kwargs):
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_task_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.get_task_with_http_info(**kwargs)  # noqa: E501
            return data

    def get_task_with_http_info(self, **kwargs):
        all_params = ['async_req', '_return_http_data_only', '_preload_content', '_request_timeout', 'uuid']
        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method task_get" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('GET', f'/task/uuid/{kwargs["uuid"]}')

    def update_task(self, data, **kwargs):
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.update_task_with_http_info(data, **kwargs)  # noqa: E501
        else:
            (data) = self.update_task_with_http_info(data, **kwargs)  # noqa: E501
            return data

    def update_task_with_http_info(self, data, **kwargs):
        all_params = ['async_req', '_return_http_data_only', '_preload_content', '_request_timeout', 'uuid', 'state']
        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method task_update" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('PUT', f'/task/uuid/{kwargs["uuid"]}/{kwargs["state"]}', body_params=data)

    def delete_task(self, **kwargs):
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.delete_task_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.delete_task_with_http_info(**kwargs)  # noqa: E501
            return data

    def delete_task_with_http_info(self, **kwargs):
        all_params = ['async_req', '_return_http_data_only', '_preload_content', '_request_timeout', 'uuid']
        params = locals()
        for key, val in params['kwargs'].items():
            if key not in all_params:
                raise TypeError("Got an unexpected keyword argument '%s'"
                                " to method profile_get" % key)
            params[key] = val
        del params['kwargs']

        return self.client.request('DELETE', f'/task/uuid/{kwargs["uuid"]}')



#!/usr/bin/env python3
# coding: utf-8
"""
    SENSE-SiteRM Debug API
"""
from sense.client.siterm.requestwrapper import RequestWrapper


class DebugApi:
    """Debug API for SENSE-SiteRM"""
    def __init__(self):
        self.client = RequestWrapper()

    def _validate_input(self, keys, kwargs, action):
        """Validate input for debug actions"""
        sitename = kwargs.get("sitename", None)
        if not sitename:
            raise Exception(f"Sitename is required for {action} test")
        if set(kwargs.keys()) != set(keys):
            raise Exception(f"Wrong input for {action}. Must have these keys: {keys}. Input: {kwargs}")
        return sitename

    def __submit_call(self, keys, **kwargs):
        """Submit a debug call to the SENSE-SiteRM API"""
        sitename = self._validate_input(keys, kwargs, kwargs.get("type"))

        return self.client.makeRequest(sitename=sitename, url=f"{kwargs['sitename']}/sitefe/json/frontend/submitdebug/NEW",
                                       **{"verb": "POST", "data": kwargs})

    def submit_ping(self, **kwargs):
        """Submit a ping test to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "ip",
                "packetsize", "interval", "interface", "time", "onetime"]
        kwargs["type"] = "rapid-ping"  # Default to rapid-ping
        return self.__submit_call(keys, **kwargs)

    def submit_arptable(self, **kwargs):
        """Submit an ARP table request to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "interface"]
        kwargs["type"] = "arp-table"  # Default to arp-table
        return self.__submit_call(keys, **kwargs)

    def submit_tcpdump(self, **kwargs):
        """Submit a tcpdump request to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "interface"]
        kwargs["type"] = "tcpdump"  # Default to tcpdump
        return self.__submit_call(keys, **kwargs)

    def submit_traceroute(self, **kwargs):
        """Submit a traceroute request to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "ip", "from_ip", "from_interface"]
        kwargs["type"] = "traceroute"  # Default to traceroute
        return self.__submit_call(keys, **kwargs)

    def submit_iperfclient(self, **kwargs):
        """Submit an iperf client request to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "interface", "ip", "port", "time"]
        kwargs["type"] = "iperf-client"  # Default to iperf-client
        return self.__submit_call(keys, **kwargs)

    def submit_iperfserver(self, **kwargs):
        """Submit an iperf server request to the SENSE-SiteRM API"""
        keys = ["hostname", "type", "sitename", "interface", "port", "time", "onetime", "ip"]
        kwargs["type"] = "iperf-server"  # Default to iperf-server
        return self.__submit_call(keys, **kwargs)

    def submit_prompush(self, **kwargs):
        """Submit a prometheus push request to the SENSE-SiteRM API"""
        keys = ["hostname", "hosttype", "type", "sitename", "gateway", "runtime", "resolution", "metadata"]
        kwargs["type"] = "prometheus-push"  # Default to prometheus-push
        return self.__submit_call(keys, **kwargs)

    def submit_arppush(self, **kwargs):
        """Submit an ARP push request to the SENSE-SiteRM API"""
        keys = ["hostname", "hosttype", "type", "sitename", "gateway", "runtime", "resolution", "metadata"]
        kwargs["type"] = "arp-push"  # Default to prometheus-push
        return self.__submit_call(keys, **kwargs)


    def get_debug(self, **kwargs):
        """Get debug info from SENSE-SiteRM Endpoint"""
        sitename = self._validate_input(["sitename", "id"], kwargs, "get-debug")
        return self.client.makeRequest(sitename=sitename,
                                       url=f"/{sitename}/sitefe/json/frontend/getdebug/{kwargs['id']}",
                                       **{"verb": "GET", "data": {}})

    def get_all_debug_hostname(self, **kwargs):
        """Get all debug info from SENSE-SiteRM Endpoint"""
        sitename = self._validate_input(["sitename", "hostname", "state"], kwargs, "get-all-debug")
        return self.client.makeRequest(sitename=sitename,
                                       url=f"/{sitename}/sitefe/json/frontend/getalldebughostname/{kwargs['hostname']}/{kwargs['state']}",
                                       **{"verb": "GET", "data": {}})

    def delete_debug(self, **kwargs):
        """Delete debug info in SENSE-SiteRM Endpoint"""
        sitename = self._validate_input(["sitename", "id"], kwargs, "delete-debug")
        return self.client.makeRequest(sitename=sitename,
                                       url=f"/{sitename}/sitefe/json/frontend/deletedebug/{kwargs['id']}",
                                       **{"verb": "DELETE", "data": {}})



# coding: utf-8

"""
    SENSE-O Northbound Intent API

    StackV SENSE-O Northbound REST API Documentation  # noqa: E501

    OpenAPI spec version: 2.0.2
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class DomainDescriptionEdgePoints(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'stp': 'Terminal',
        'capailities': 'list[CapabilityDescription]',
        'peer_uri': 'str',
        'peer_name': 'str'
    }

    attribute_map = {
        'stp': 'stp',
        'capailities': 'capailities',
        'peer_uri': 'peer_uri',
        'peer_name': 'peer_name'
    }

    def __init__(self, stp=None, capailities=None, peer_uri=None, peer_name=None):  # noqa: E501
        """DomainDescriptionEdgePoints - a model defined in Swagger"""  # noqa: E501
        self._stp = None
        self._capailities = None
        self._peer_uri = None
        self._peer_name = None
        self.discriminator = None
        if stp is not None:
            self.stp = stp
        if capailities is not None:
            self.capailities = capailities
        if peer_uri is not None:
            self.peer_uri = peer_uri
        if peer_name is not None:
            self.peer_name = peer_name

    @property
    def stp(self):
        """Gets the stp of this DomainDescriptionEdgePoints.  # noqa: E501


        :return: The stp of this DomainDescriptionEdgePoints.  # noqa: E501
        :rtype: Terminal
        """
        return self._stp

    @stp.setter
    def stp(self, stp):
        """Sets the stp of this DomainDescriptionEdgePoints.


        :param stp: The stp of this DomainDescriptionEdgePoints.  # noqa: E501
        :type: Terminal
        """

        self._stp = stp

    @property
    def capailities(self):
        """Gets the capailities of this DomainDescriptionEdgePoints.  # noqa: E501


        :return: The capailities of this DomainDescriptionEdgePoints.  # noqa: E501
        :rtype: list[CapabilityDescription]
        """
        return self._capailities

    @capailities.setter
    def capailities(self, capailities):
        """Sets the capailities of this DomainDescriptionEdgePoints.


        :param capailities: The capailities of this DomainDescriptionEdgePoints.  # noqa: E501
        :type: list[CapabilityDescription]
        """

        self._capailities = capailities

    @property
    def peer_uri(self):
        """Gets the peer_uri of this DomainDescriptionEdgePoints.  # noqa: E501


        :return: The peer_uri of this DomainDescriptionEdgePoints.  # noqa: E501
        :rtype: str
        """
        return self._peer_uri

    @peer_uri.setter
    def peer_uri(self, peer_uri):
        """Sets the peer_uri of this DomainDescriptionEdgePoints.


        :param peer_uri: The peer_uri of this DomainDescriptionEdgePoints.  # noqa: E501
        :type: str
        """

        self._peer_uri = peer_uri

    @property
    def peer_name(self):
        """Gets the peer_name of this DomainDescriptionEdgePoints.  # noqa: E501


        :return: The peer_name of this DomainDescriptionEdgePoints.  # noqa: E501
        :rtype: str
        """
        return self._peer_name

    @peer_name.setter
    def peer_name(self, peer_name):
        """Sets the peer_name of this DomainDescriptionEdgePoints.


        :param peer_name: The peer_name of this DomainDescriptionEdgePoints.  # noqa: E501
        :type: str
        """

        self._peer_name = peer_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(DomainDescriptionEdgePoints, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DomainDescriptionEdgePoints):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



from typing import List

from sense.workflow.base.config_models import DependencyInfo
from sense.workflow.base.constants import Constants
from sense.workflow.base.utils import get_logger
from sense.workflow.provider.dependency_util import get_values_for_dependency
from .sense_service import SenseService

logger = get_logger()


class ServiceHandler:
    def __init__(self, *, provider):
        self.provider = provider
        self._being_deleted = dict()

    @property
    def client(self):
        return self.provider.client

    @property
    def name(self):
        return self.provider.name

    @property
    def resource_listener(self):
        return self.provider.resource_listener

    @property
    def services(self) -> List[SenseService]:
        services: List[SenseService] = list()

        for serv in self.provider.services:
            if isinstance(serv, SenseService):
                services.append(serv)

        return services

    @staticmethod
    def _load_manifest_template(config_dir, manifest_template) -> dict:
        import json
        from pathlib import Path
        import os
        import yaml

        if not Path(manifest_template).is_absolute():
            manifest_template = os.path.join(config_dir, manifest_template)

        if manifest_template.endswith('.json'):
            with open(manifest_template, 'r') as fp:
                manifest_template = json.load(fp)
        else:
            with open(manifest_template, 'r') as stream:
                manifest_template = yaml.safe_load(stream)

        return manifest_template

    def add_resource(self, *, resource: dict):
        assert Constants.PROFILE in resource
        label = resource[Constants.LABEL]
        profile = resource[Constants.PROFILE]
        assert profile, f"Must have a profile for {label}"
        edit_template = resource.get(Constants.EDIT_TEMPLATE, dict())
        count = resource[Constants.RES_COUNT]
        manifest_template = resource.get("manifest_template", dict())

        if isinstance(manifest_template, str):
            manifest_template = self._load_manifest_template(resource.get(Constants.CONFIG_DIR), manifest_template)

        for idx in range(0, count):
            serv_name = self.provider.resource_name(resource, idx)
            saved_state = next(filter(lambda s: s.attributes['name'] == serv_name, resource[Constants.SAVED_STATES]),
                               dict())
            serv = SenseService(client=self.client, label=label, name=serv_name, profile=profile,
                                edit_template=edit_template,
                                manifest_template=manifest_template,
                                saved_state=saved_state
                                )
            self.provider.services.append(serv)
            self.resource_listener.on_added(source=self.provider, provider=self.provider, resource=serv)

    def create_resource(self, *, resource: dict):
        label = resource[Constants.LABEL]

        if self.provider.modified:
            for serv_name in set(self.provider.existing_map[label]) - set(self.provider.added_map[label]):
                logger.warning(f"{self.name}:Modified count of resource: Deleting {serv_name} ...")
                saved_state = next(filter(lambda s: s.attributes['name'] == serv_name,
                                          resource[Constants.SAVED_STATES]))
                profile = saved_state.attributes[Constants.PROFILE]
                edit_template = saved_state.attributes["edit_template"]
                manifest_template = saved_state.attributes["manifest_template"]
                serv = SenseService(
                    client=self.client,
                    label=label, name=serv_name, profile=profile,
                    edit_template=edit_template,
                    manifest_template=manifest_template)

                serv.delete()
                self._being_deleted[serv_name] = serv
                logger.warning(f"Done deleting sense resource:{serv_name}")

        if Constants.PROFILE not in resource:
            return

        serv = next(filter(lambda s: s.label == label, self.services))

        for k, v in serv.edit_template.items():
            if isinstance(v, DependencyInfo):
                values = get_values_for_dependency(resource=resource,
                                                   attribute=Constants.EDIT_TEMPLATE + "." + k)
                serv.edit_template[k] = values[0]

        logger.debug(f"{self.name}: Creating resource with label {label} ....")

        for serv in filter(lambda s: s.label == label, self.services):
            logger.info(f"Creating resource: {serv.name}")
            serv.create()

    def wait_for_create_resource(self, *, resource: dict):
        label = resource[Constants.LABEL]
        for serv_name, serv in self._being_deleted.copy().items():
            if serv.label != label:
                continue

            logger.info(f"{self.name}: Checking modified in wait_for_create_resource: {serv_name} ....")
            self._being_deleted.pop(serv_name)
            serv.wait_for_delete()
            logger.warning(f"Done waiting on deleting sense resource:{serv_name}")
            self.resource_listener.on_deleted(source=self.provider, provider=self.provider, resource=serv)

        for serv in filter(lambda s: s.label == label, self.services):
            logger.debug(f"Waiting on Create resource: {serv.name}")
            serv.wait_for_create()
            logger.debug(f"Resource has been created {serv.name}")
            self.resource_listener.on_created(source=self.provider, provider=self.provider, resource=serv)
            logger.debug(f"Notified Resource has been created: {serv.name}")

    def delete_resource(self, *, resource: dict):
        assert Constants.PROFILE in resource

        label = resource[Constants.LABEL]
        logger.info(f"Deleting resource: {label}")
        edit_template = resource.get("edit_template", dict())
        profile = resource[Constants.PROFILE]
        count = resource[Constants.RES_COUNT]
        manifest_template = resource.get("manifest_template", dict())

        if isinstance(manifest_template, str):
            manifest_template = self._load_manifest_template(resource.get(Constants.CONFIG_DIR), manifest_template)

        services = list()

        for idx in range(0, count):
            serv_name = self.provider.resource_name(resource, idx)
            serv = SenseService(client=self.client,
                                label=label, name=serv_name,
                                profile=profile,
                                edit_template=edit_template,
                                manifest_template=manifest_template)

            serv.delete()
            services.append(serv)

        for serv in services:
            serv.wait_for_delete()
            self.resource_listener.on_deleted(source=self.provider, provider=self.provider, resource=serv)

        logger.info(f"Done Deleting resource: {label}")



import os
import unittest
import pytest
import re
import json
import time
import sys

# Append root for proper imports
sys.path.append('')
#

#
from sense.client.address_api import AddressApi
from sense.common import loadJSON
#


class TestAddressApi(unittest.TestCase):

    def setUp(self) -> None:
        self.client = AddressApi()
        self.pool = 'IPv6-Pool-T1'
        self.scope = 'TestScope1'
        # TODO: create test pool update setUp

    def tearDown(self) -> None:
        # TODO: delet test pool update tearDown
        pass

    def test_get_pool(self):
        # - TESTING: GET /address/allocate
        res = self.client.get_allocations(self.pool)
        print(res)

    def test_get_pool_scope(self):
        # - TESTING: GET /address/allocate/{scope}
        res = self.client.get_allocations(self.pool, scope=self.scope)
        print(res)

    def test_allocate_ipv6(self):
        # - TESTING: POST /address/allocate/{pool_name}
        res = self.client.allocate_ipv4_address(self.pool, 'alloc1', batch=3, scope=self.scope)
        print(res)
        res = self.client.allocate_ipv4_address_subnet(self.pool, 'alloc2', batch = 3, netmask='/50', scope=self.scope)
        print(res)
        res = self.client.allocate_ipv4_subnet(self.pool, 'alloc3', netmask='/64', scope=self.scope)
        print(res)


    def test_free_ipv6(self):
        # - TESTING: DELETE /address/allocate/{pool_name}/...
        #res = self.client.free_address(self.pool, name='alloc2', scope=self.scope)
        res = self.client.free_address(self.pool, scope=self.scope, address='2001:48d0:3001:8000::/50')
        print(res)


    def test_affiliate_ipv6(self):
        # - TESTING: POST /address/affiliate/{pool_name}/...
        # res = self.client.affiliate_address(self.pool, 'uri:test:test', scope=self.scope, address='2001:48d0:3001::/50')
        res = self.client.affiliate_address(self.pool, 'uri:test:test2', scope=self.scope, name='2001:48d0:3001::/50')
        print(res)





#!/usr/bin/env python3
# coding: utf-8
"""
    SENSE-SiteRM Models API
"""
from sense.client.siterm.requestwrapper import RequestWrapper


class ModelsApi:
    """Models API for SENSE-SiteRM"""
    def __init__(self):
        self.client = RequestWrapper()

    def _getSitename(self, **kwargs):
        sitename = kwargs.get("sitename", None)
        if not sitename:
            raise Exception("Sitename is required for get_configuration")
        return sitename

    def get_models(self, **kwargs):
        """Get models from the SENSE-SiteRM API"""
        sitename = self._getSitename(**kwargs)
        # kwargs['urlparams'] = {'encode': 'false', 'summary': 'false', 'oldview': 'true',
        #                        'current': 'False', 'model': 'turtle'}
        for key in ['encode', 'summary', 'oldview', 'current', 'model']:
            if key in kwargs:
                kwargs.setdefault('urlparams', {})
                kwargs['urlparams'][key] = kwargs[key]
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/models",
                                        **{"verb": "GET", "data": {}, "urlparams": kwargs.get('urlparams', None)})

    def get_model(self, **kwargs):
        """Get model from the SENSE-SiteRM API"""
        sitename = self._getSitename(**kwargs)
        # kwargs['urlparams'] = {'encode': 'false', 'summary': 'false'}
        for key in ['encode', 'summary']:
            if key in kwargs:
                kwargs.setdefault('urlparams', {})
                kwargs['urlparams'][key] = kwargs[key]
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/models/{kwargs.get('model_id')}",
                                        **{"verb": "GET", "data": {}})



#!/usr/bin/env python3
# coding: utf-8
"""
    SENSE-SiteRM Deltas API
"""
from sense.client.siterm.requestwrapper import RequestWrapper


class DeltasApi:
    """Deltas API for SENSE-SiteRM"""
    def __init__(self):
        self.client = RequestWrapper()

    def _getSitename(self, **kwargs):
        sitename = kwargs.get("sitename", None)
        if not sitename:
            raise Exception("Sitename is required for get_configuration")
        return sitename

    def get_deltas(self, **kwargs):
        """Get deltas from the SENSE-SiteRM API"""
        sitename = self._getSitename(**kwargs)
        # kwargs['urlparams'] = {'encode': 'false', 'summary': 'false', 'oldview': 'true', 'model': 'turtle'}
        for key in ['encode', 'summary', 'oldview', 'model']:
            if key in kwargs:
                kwargs.setdefault('urlparams', {})
                kwargs['urlparams'][key] = kwargs[key]
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/deltas",
                                        **{"verb": "GET", "data": {}, "urlparams": kwargs.get('urlparams', None)})

    def get_delta(self, **kwargs):
        """Get delta from the SENSE-SiteRM API"""
        sitename = self._getSitename(**kwargs)
        for key in ['encode', 'summary', 'oldview', 'model']:
            if key in kwargs:
                kwargs.setdefault('urlparams', {})
                kwargs['urlparams'][key] = kwargs[key]
        # kwargs['urlparams'] = {'encode': 'false', 'summary': 'false', 'oldview': 'true', 'model': 'turtle'}
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/deltas/{kwargs.get('delta_id')}",
                                        **{"verb": "GET", "data": {}, "urlparams": kwargs.get('urlparams', None)})

    def get_delta_states(self, **kwargs):
        """Get delta states from the SENSE-SiteRM API"""
        sitename = self._getSitename(**kwargs)
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/deltastates/{kwargs.get('delta_id')}",
                                        **{"verb": "GET", "data": {}})


    def forceapplydelta(self, **kwargs):
        """Force Apply Delta inside SiteRM"""
        sitename = self._getSitename(**kwargs)
        if not ('uuid' in kwargs and kwargs['uuid']):
            raise Exception("Force apply requires delta uuid to be passed as argument.")
        return self.client.makeRequest(sitename=sitename,
                                        url=f"/{sitename}/sitefe/v1/deltaforceapply",
                                        **{"verb": "POST", "data": {"uuid": kwargs["uuid"]}})



import setuptools

from sense import __VERSION__

with open("requirements.txt", "r", encoding="utf-8") as fh:
  requirements = fh.read()

setuptools.setup(
  name="sense_o_api",
  version=__VERSION__,
  author="Xi Yang",
  description="SENSE-Orchestrator Northbound API Client",
  url="https://github.com/esnet/StackV",
  long_description="SENSE-Orchestrator Northbound API Client Library",
  long_description_content_type="text/plain",
  packages=setuptools.find_packages(),
  include_package_data=True,
  scripts=['util/sense_util.py', 'sense/workflow/sense_workflow.py'],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires=">=3.6",
  install_requires=requirements,
  setup_requires=requirements,
)



import enum

AUTH_ENDPOINT = "auth_endpoint"
API_ENDPOINT = "api_endpoint"
CLIENT_ID = "client_id"
USERNAME = "username"
PASSWORD = "password"
SECRET = "secret"

SENSE_CONF_ATTRS = [AUTH_ENDPOINT, API_ENDPOINT, CLIENT_ID, USERNAME, PASSWORD, SECRET]

SENSE_PROFILE_UID = "service_profile_uuid"
SENSE_ALIAS = "alias"
SENSE_EDIT = "options"
SENSE_URI = 'uri'
# SENSE_ID = 'id'
SENSE_PATH = 'path'
SENSE_VLAN_TAG = 'vlan_tag'


SERVICE_INSTANCE_KEYS = ['intents', 'referenceUUID', 'state']

SENSE_CUSTOMER_ASN = "customer_asn"
SENSE_AMAZON_ASN = "amazon_asn"
SENSE_GOOGLE_ASN = "google_asn"
SENSE_CUSTOMER_IP = "customer_ip"
SENSE_AMAZON_IP = "amazon_ip"
SENSE_GOOGLE_IP = "google_ip"
SENSE_AUTHKEY = "authkey"
SENSE_TO_HOSTED_CONN = "to_hosted_conn"

SENSE_GCP_PEERING_MAPPING = {
    'local_asn': SENSE_CUSTOMER_ASN,
    'local_address': SENSE_CUSTOMER_IP,
    'remote_asn': SENSE_GOOGLE_ASN,
    'remote_address': SENSE_GOOGLE_IP,
    'RES_SECURITY': SENSE_AUTHKEY
}

SENSE_KEYPAIR = 'Key Pair'
SENSE_PUBLIC_IP = 'Public IP'
SENSE_PRIVATE_IP = 'Private IP'
SENSE_NODE_NAME = 'Node Name'
SENSE_IMAGE = 'Image'

SENSE_RETRY = 50


class SupportedCloud(str, enum.Enum):
    """
    The Cloud supported by SENSE.
    """
    GCP = "GOOGLE"



import os
import unittest
import pytest
import re
import json
import string
import random
import sys

# Append root for proper imports
sys.path.append('')
#

#
from sense.client.intent_api import IntentApi
from sense.client.instance_api import InstanceApi
from sense.common import loadJSON
#


class TestIntentApi(unittest.TestCase):
    def setUp(self) -> None:
        self.client = IntentApi()

        self.instance_client = InstanceApi()
        self.instance_client.instance_new()
        input = loadJSON("requests/request-1.json")
        self.instance_client.instance_create(json.dumps(input))

        self.client.si_uuid = self.instance_client.si_uuid

    def tearDown(self) -> None:
        if self.instance_client.si_uuid is not None:
            self.instance_client.instance_delete()

    def test_get_intents(self):
        #
        # TESTING: - /intent/instance/{siUUID}
        res = json.loads(self.client.instance_get_intents())
        assert len(res) > 0 and res[0].get("id")

    def test_get_intent_id(self):
        # PRE-OP: get uuid from test instance.
        uuid = json.loads(self.client.instance_get_intents())[0].get("id")

        #
        # TESTING: - /intent/{uuid}
        res = json.loads(self.client.intent_describe(uuid))
        assert res.get("id") == uuid


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from .provider import Provider


class ResourceListener(ABC):
    @abstractmethod
    def on_added(self, *, source, provider: Provider, resource: object):
        pass

    @abstractmethod
    def on_created(self, *, source, provider: Provider, resource: object):
        pass

    @abstractmethod
    def on_deleted(self, *, source, provider: Provider, resource: object):
        pass
